"""
.. codeauthor:: Rich Plevin <rich@plevin.com>

.. Copyright (c) 2020 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
"""
from ..subcommand import SubcommandABC, clean_help
from ..log import getLogger

_logger = getLogger(__name__)

def get_re_techs(tech_cols, params, region):
    return [tech for tech in tech_cols if params.loc[region, tech] == 1]

def element_path(elt, withInput):
    d = {'input' : elt.attrib['name']}

    sector = subsector = technology = None

    for node in elt.iterancestors():    # walk up the hierarchy
        tag = node.tag
        attr = node.attrib

        if tag == 'period':
            pass

        elif tag == 'location-info':
            d['sector'] = sector = attr['sector-name']
            d['subsector'] = subsector = attr['subsector-name']

        elif tag == 'supplysector':
            d['sector'] = sector = attr['name']

        elif tag in ('stub-technology', 'technology'):
            d['technology'] = technology = attr['name']

        elif tag == 'subsector':
            d['subsector'] = subsector = attr['name']

        elif tag in ('global-technology-database', 'region'):
            break

    return (sector, subsector, technology, d['input']) if withInput else (sector, subsector, technology)



def save_bldg_techs(f, args, years, xml_file, xpath, which, withInput):
    from ..config import getParam
    from ..utils import pathjoin
    from ..XMLFile import XMLFile

    gcamDir = getParam('GCAM.RefWorkspace', section=args.projectName)
    pathname = pathjoin(gcamDir, 'input', 'gcamdata', 'xml', xml_file)

    _logger.info("Reading {}".format(pathname))
    xml = XMLFile(pathname)
    root = xml.getRoot()

    nodes = root.xpath(xpath)
    paths = sorted(set([element_path(node, withInput) for node in nodes])) # use 'set' to remove dupes

    # filter out sectors missing from cmdline arg, if specified
    if args.sectors:
        desired = []
        sectors = set(args.sectors.split(','))
        for path in paths:
            if path[0] in sectors:
                desired.append(path)
        paths = desired

    all_regions = set(root.xpath('//region/@name'))
    if args.GCAM_USA:
        all_regions = all_regions.difference(['USA'])  # remove USA since states will be used

    regions = args.regions.split(',') if args.regions else all_regions
    regions = sorted(regions)

    zeroes = ',0' * len(years)    # fill in with zeroes for reading into a dataframe

    # data values
    for region in regions:
        if region not in all_regions:   # use only regions defined for this XML file
            continue

        for tup in paths:
            f.write(which + ',' + region + ',')
            f.write(','.join(tup))
            f.write(zeroes + '\n')

TECH_OUTPUT_FILE = 'industry_tech_template.csv'
ELEC_OUTPUT_FILE = 'industry_elec_template.csv'

class IndustryCommand(SubcommandABC):

    def __init__(self, subparsers):
        kwargs = {'help' : '''Dump combinations of industry energy use sectors, techs, and fuels.'''}
        super(IndustryCommand, self).__init__('industry', subparsers, kwargs, group='project')

    def addArgs(self, parser):
        parser.add_argument('-e', '--electricOnly', action="store_true",
                            help=clean_help('''Generate a template for electricity-based sources only.'''))


        parser.add_argument('-o', '--outputFile', default=None,
                            help=clean_help('''The CSV file to create with lists of unique industry sectors, 
                            subsectors, and technologies. Default is "[GCAM.CsvTemplateDir]/{}" or 
                            "[GCAM.CsvTemplateDir]/{}", depending on the --electricOnly flag.
                            Use an absolute path to generate the file to another location.'''.format(
                                TECH_OUTPUT_FILE, ELEC_OUTPUT_FILE)))

        parser.add_argument('-s', '--sectors', default="industrial energy use",
                            help=clean_help('''A comma-delimited list of sectors to include in the generated template. Use quotes 
                            around the argument if there are embedded blanks. By default, all known industry technology
                            sectors are included.'''))

        parser.add_argument('-r', '--regions', default=None,
                            help=clean_help('''A comma-delimited list of regions to include in the generated template. 
                             By default all regions are included. '''))

        parser.add_argument('-u', '--GCAM-USA', action="store_true",
                            help=clean_help('''If set, produce output compatible with GCAM-USA regions.'''))

        parser.add_argument('-y', '--years', default='2015-2100',
                            help=clean_help('''A hyphen-separated range of timestep years to include in the generated template.
                            Default is "2015-2100"'''))

        return parser


    def run(self, args, tool):
        from ..utils import pathjoin, validate_years
        from ..config import getParam

        outputFile = args.outputFile or (ELEC_OUTPUT_FILE if args.electricOnly else TECH_OUTPUT_FILE)

        main_xml_file = 'industry.xml'
        usa_xml_file = 'industry_USA.xml'

        main_xpath = '//supplysector/subsector/stub-technology/period/minicam-energy-input'
        usa_xpath = '//supplysector/subsector/stub-technology/period/minicam-energy-input'

        templateDir = getParam('GCAM.CsvTemplateDir')
        outputPath = pathjoin(templateDir, outputFile)

        _logger.info('Writing {}'.format(outputPath))
        with open(outputPath, 'w') as f:
            years = validate_years(args.years)
            if years is None:
                raise Exception(
                    'Year argument must be two integers separated by a hyphen, with second > first. Got "{}"'.format(
                        args.years))

            # column headers
            f.write("which,region,sector,subsector,technology,")
            withInput = not args.electricOnly
            if withInput:
                f.write("input,")

            f.write(','.join(map(str, years)))
            f.write("\n")

            save_bldg_techs(f, args, years, main_xml_file, main_xpath, 'GCAM-32', withInput)

            if args.GCAM_USA:
                save_bldg_techs(f, args, years, usa_xml_file, usa_xpath, 'GCAM-USA', withInput)



from lxml import etree as ET
from .config import pathjoin, getParam
from .log import getLogger
from .XMLFile import XMLFile

_logger = getLogger(__name__)

def element_path(elt):
    d = {'technology' : elt.attrib['name']}

    for node in elt.iterancestors():    # walk up the hierarchy
        tag = node.tag
        attr = node.attrib

        if tag == 'region':
            d['region'] = attr['name' ]
            break

        # elif tag == 'location-info':
        #     d['sector'] = attr['sector-name']
        #     d['subsector'] = attr['subsector-name']

        elif tag == 'supplysector':
            d['sector'] = attr['name']

        elif tag == 'tranSubsector':
            d['subsector'] = attr['name']

    return (d['region'], d['sector'], d['subsector'], d['technology'])


def zevPolicyMain(args):
    import pandas as pd
    from .error import CommandlineError
    from .xmlSetup import scenarioXML
    from .utils import validate_years, get_path

    years = validate_years(args.years)
    if years is None:
        raise CommandlineError(
            'Year argument must be two integers separated by a hyphen, with second > first. Got "{}"'.format(
                args.years))

    outputCSV  = args.outputCSV
    useGcamUSA = args.GCAM_USA

    outPath = get_path(outputCSV, pathjoin(getParam("GCAM.ProjectDir"), "etc"))

    # parse sect:subsect:tech args
    tuples = []
    for arg in args.include:
        # adding 2 colons ensures that split() returns at least 3 substrings
        (sectors, subsects, techs) = (arg + "::").split(':')[0:3]
        tuples.append((sectors.split(','), subsects.split(','), techs.split(',')))

    def match_expr(names):
        expr = " or ".join(['@name = "{}"'.format(name) for name in names if name])
        return ('[' + expr +']') if expr else ''

    xpaths =[]
    for (sectors, subsects, techs)  in tuples:
        sector_match  = match_expr(sectors)
        subsect_match = match_expr(subsects)
        tech_match    = match_expr(techs)
        xpath = '/supplysector{}/tranSubsector{}/stub-technology{}'.format(sector_match, subsect_match, tech_match)
        xpaths.append(xpath)

    transportXML = scenarioXML(args.scenario, args.tag) # read the file associated with the given tag
    xml = XMLFile(transportXML)
    root = xml.getRoot()

    if args.regions:
        regions = args.regions.split(',')
        comps = ['@name="{}"'.format(r) for r in regions]
        regionPrefix = '//region[{}]'.format(' or '.join(comps))
    else:
        regionPrefix = '//region'

    techs = []
    combos = dict()
    for xpath in xpaths:
        full_xpath = regionPrefix + xpath

        nodes = root.xpath(full_xpath)
        techs += [node.attrib['name'] for node in nodes]

        combos.update({element_path(node) : True for node in nodes})

    unique_techs = sorted(set(techs))
    print("techs:", unique_techs)

    # create list of tuples without techs
    active_subsects = [list(tup) for tup in sorted(set([key[0:3] for key in combos.keys()]))]

    year_cols = [str(year) for year in years]

    # generate CSV template
    columns = ['region', 'market', 'supplysector', 'tranSubsector'] + unique_techs + year_cols
    df = pd.DataFrame(columns=columns, index=range(0, len(active_subsects)))

    df[year_cols] = 0

    # convert row tuples to columns to update the df
    all_regs, all_sects, all_subsects = zip(*active_subsects)
    df['market'] = all_regs     # default market is same as region name
    df['region'] = all_regs
    df['supplysector'] = all_sects
    df['tranSubsector'] = all_subsects

    for idx, row in df.iterrows():
        for tech in unique_techs:
            key = (row.region, row.supplysector, row.tranSubsector, tech)
            if combos.get(key, False):
               df.loc[idx, tech] = 1

    _logger.info("Writing '%s'", outPath)
    df.to_csv(outPath, index=None)


BtuToMJ = 1.055

def generate_zev_xml(scenario, csvPath, xmlPath, transportTag, pMultiplier, outputRatio):
    import os
    import pandas as pd
    from .utils import mkdirs
    from .RESPolicy import write_xml
    from .xmlSetup import scenarioXML

    df = pd.read_csv(csvPath, index_col=None)

    common_cols = ['region', 'market', 'supplysector', 'tranSubsector']
    year_cols = [col for col in df.columns if col.isdigit()]
    tech_cols = list(set(df.columns) - set(common_cols + year_cols))

    def set_text(elt, value):
        elt.text = str(value)
        return elt

    def rec_name(region, subsector, year, policy='ZEC'):
        name = '{} {} {} {}'.format(region, subsector, policy, year)
        return name

    # we use the indicated transportation XML file to extract load factors
    transportXML = scenarioXML(scenario, transportTag) # read the file associated with the given tag
    xml = XMLFile(transportXML)
    trans_root = xml.getRoot()

    def load_factor(region, sector, subsector, tech, year):
        xpath = "//region[@name='{}']/supplysector[@name='{}']/tranSubsector[@name='{}']/stub-technology[@name='{}']/period[@year='{}']/loadFactor".format(
            region, sector, subsector, tech, year)
        nodes = trans_root.xpath(xpath)
        if len(nodes) == 0:
            raise Exception('ZEVPolicy: Failed to find loadFactor for "{}"'.format(xpath))

        return float(nodes[0].text)

    def find_or_create(parent, tag, name):
        elt = parent.find('{}[@name="{}"]'.format(tag, name))
        if elt is None:
            elt = ET.SubElement(parent, tag, name=name)

        return elt

    root = ET.Element('scenario')
    world = ET.SubElement(root, 'world')

    # If the standard is zero in some years, emit no policy-portfolio-standard, no
    # coefficients, and no secondary-res-outputs for that year.

    for region in sorted(df.region.unique()):
        region_df = df.query('region == @region')

        region_elt = find_or_create(world, 'region', region)

        for idx, row in region_df.iterrows():
            sector    = row.supplysector
            subsector = row.tranSubsector
            market    = row.market

            # <policy-portfolio-standard name="{region} {tranSubsector} ZEC {year}">
            #   <market>{market}</market>
            #   <policyType>RES</policyType>
            #   <constraint year="{year}">1</constraint>
            # </policy-portfolio-standard>
            for year in year_cols:
                target = region_df.loc[idx, year]
                if target:
                    name = rec_name(market, subsector, year)
                    std_elt = ET.SubElement(region_elt, 'policy-portfolio-standard', name=name)
                    set_text(ET.SubElement(std_elt, 'market'), market)
                    set_text(ET.SubElement(std_elt, 'policyType'), 'RES')
                    set_text(ET.SubElement(std_elt, 'constraint', year=year), 1)

            sect_elt  = None  # created on demand in loop below

            # <supplysector name="trn_freight_road">
            #   <tranSubsector name="Truck (>32t)">
            #     <stub-technology name="electric">
            #       <period year="2030">
            #         <minicam-energy-input name="EU15 Truck (>32t) ZEC 2030">
            #           <coefficient>1544.0758</coefficient>
            #         </minicam-energy-input>
            #         <res-secondary-output name="EU15 Truck (>32t) ZEC 2030">
            #           <output-ratio>0.000001</output-ratio>
            #           <pMultiplier>1.25E10</pMultiplier>
            #         </res-secondary-output>
            #       </period>
            for tech in tech_cols:
                for year in year_cols:
                    target = region_df.loc[idx, year]
                    if not target:
                        continue

                    sect_elt = sect_elt if sect_elt is not None else find_or_create(region_elt, 'supplysector', sector)
                    subsect_elt = find_or_create(sect_elt, 'tranSubsector', subsector)

                    tech_elt = find_or_create(subsect_elt, 'stub-technology', tech)
                    period_elt = ET.SubElement(tech_elt, 'period', year=year)
                    name = rec_name(market, subsector, year)

                    en_input_elt = ET.SubElement(period_elt, 'minicam-energy-input', name=name)

                    value = target / BtuToMJ * 1E3 * load_factor(region, sector, subsector, tech, year)
                    set_text(ET.SubElement(en_input_elt, 'coefficient'), value)

                    # create res-secondary-output elements only for technologies included in the policy
                    if region_df.loc[idx, tech]:
                        sec_input_elt = ET.SubElement(period_elt, 'res-secondary-output', name=name)
                        set_text(ET.SubElement(sec_input_elt, 'output-ratio'), outputRatio)
                        set_text(ET.SubElement(sec_input_elt, 'pMultiplier'), pMultiplier)

    # delete empty regions
    emptyRegs = [reg for reg in world if len(reg) == 0]
    for reg in emptyRegs:
        world.remove(reg)

    _logger.info("Writing '%s'", xmlPath)
    mkdirs(os.path.dirname(xmlPath))    # ensure the location exists
    tree = ET.ElementTree(root)
    write_xml(tree, xmlPath)



# Copyright (c) 2016  Richard Plevin
# See the https://opensource.org/licenses/MIT for license details.

from pygcam.log import getLogger
from .McsSubcommandABC import McsSubcommandABC, clean_help

_logger = getLogger(__name__)

def driver(args, tool):
    """
    Run a command for each trialDir or scenarioDir, using str.format to pass
    required args. Possible format args are: projectName, simId, trialNum,
    scenario, simDir, trialDir, and scenarioDir.
    """
    from subprocess import call
    from six.moves import xrange
    from pygcam.config import getSection

    from ..Database import getDatabase
    from ..error import PygcamMcsUserError
    from ..context import Context
    from .. import util as U

    simId    = args.simId
    command  = args.command
    scenario = args.scenario
    noRun    = args.noRun
    trialStr = args.trials

    projectName = getSection()

    if trialStr:
        trials = U.parseTrialString(trialStr)
    else:
        db = getDatabase()
        count = db.getTrialCount(simId)
        trials = xrange(count)

    # TBD: Add groupName
    context = Context(projectName=projectName, simId=simId, scenario=scenario)
    _logger.info('Running iterator for projectName=%s, simId=%d, scenario=%s, trials=%s, command="%s"',
                 projectName, simId, scenario, trialStr, command)

    # Create a dict to pass to str.format. These are constant across trials.
    argDict = {
        'projectName' : projectName,
        'simId'       : args.simId,
        'scenario'    : args.scenario,
        'expName'     : args.scenario,
    }

    for trialNum in trials:
        argDict['trialNum'] = context.trialNum = trialNum
        argDict['expDir']   = argDict['scenarioDir'] = context.getScenarioDir(create=True)
        argDict['trialDir'] = context.getTrialDir()
        argDict['simDir']   = context.getSimDir()

        try:
            cmd = command.format(**argDict)
        except Exception as e:
            raise PygcamMcsUserError("Bad command format: %s" % e)

        if noRun:
            print(cmd)
        else:
            try:
                call(cmd, shell=True)

            except Exception as e:
                raise PygcamMcsUserError("Failed to run command '%s': %s" % (cmd, e))


class IterateCommand(McsSubcommandABC):
    def __init__(self, subparsers):
        kwargs = {'help' : '''Run a command in each trialDir, or if scenario is given, 
        in each expDir. The following arguments are available for use in the command string,
        specified within curly braces: projectName, simId, trialNum, scenario, expName (a
        legacy alias for scenario), trialDir, scenarioDir, and expDir (a legacy alias for
        scenarioDir). For example, to run the fictional program "foo" in each trialDir for 
        a given set of parameters, you might write:
        gt iterate -s1 -c "foo -s{simId} -t{trialNum} -i{trialDir}/x -o{trialDir}/y/z.txt".'''}
        super(IterateCommand, self).__init__('iterate', subparsers, kwargs)

    def addArgs(self, parser):
        parser.add_argument('-c', '--command', type=str, required=True,
                            help=clean_help('''A command string to execute for each trial. The following
                            arguments are available for use in the command string, specified
                            within curly braces: projectName, simId, trialNum, scenario, expName, 
                            trialDir, expDir.'''))

        parser.add_argument('-n', '--noRun', action='store_true',
                            help=clean_help("Show the commands that would be executed, but don't run them"))

        parser.add_argument('-s', '--simId', type=int, default=1,
                            help=clean_help('The id of the simulation. Default is 1.'))

        parser.add_argument('-S', '--scenario', type=str, default="",
                            help=clean_help('The name of the scenario'))

        parser.add_argument('-t', '--trials', type=str, default=None,
                             help=clean_help('''Comma separated list of trial or ranges of trials to run. Ex: 1,4,6-10,3.
                             Defaults to running all trials for the given simulation.'''))

        return parser   # for auto-doc generation


    def run(self, args, tool):
        driver(args, tool)



#! /usr/bin/env python
'''
Created on 4/26/15

@author: rjp
'''
from ..subcommand import SubcommandABC, clean_help

DefaultName  = 'cellulosic ethanol'
DefaultTag   = 'cell-etoh'
DefaultType  = 'subsidy'

class DeltaConstraintsCommand(SubcommandABC):
    def __init__(self, subparsers):
        kwargs = {#'aliases' : ['delta'],
                  'help' : clean_help('''Specify incremental values to add to the production of a given fuel,
                              by year, and generate the corresponding constraint file.''')}

        super(DeltaConstraintsCommand, self).__init__('deltaConstraint', subparsers, kwargs)

    def addArgs(self, parser):
        from ..constraints import PolicyChoices, DefaultCellulosicCoefficients, DefaultYears

        parser.add_argument('-c', '--coefficients',
                            help=clean_help('''A comma-separated string of year:coefficient pairs. This
                            sets the cellulosic ethanol conversion coefficients. Defaults to
                            standard GCAM values: %s.''' % DefaultCellulosicCoefficients))

        parser.add_argument('-b', '--baseline', required=True,
                            help=clean_help('The baseline on which the policy scenario is based'))

        parser.add_argument('-B', '--biomassPolicyType', choices=PolicyChoices, default=None,
                            help=clean_help('Regional biomass policy type. Default is None.'))

        parser.add_argument('-f', '--fuelName', default=DefaultName,
                            help=clean_help("The fuel to generate constraints for. Default is %s" % DefaultName))

        parser.add_argument('-l', '--defaultDelta', type=float, default=0.0,
                            help=clean_help('''Default increment to add to each year (EJ). All or individual
                            years values can be set (overriding -l flag values) using the -L flag.'''))

        parser.add_argument('-L', '--annualDeltas', default='',
                            help=clean_help('''Optional production increments by year. Value must be a
                            comma-delimited string of year:level pairs, where level in is EJ.
                            If -l is not used to set default for all years, you must specify
                            values for all years using this option.'''))

        # parser.add_argument('-m', '--fromMCS', action='store_true',
        #                      help=clean_help("Used when calling from gcammcs so correct pathnames are computed."))

        parser.add_argument('-p', '--policy', required=True,
                            help=clean_help('The policy scenario name'))

        parser.add_argument('-P', '--purposeGrownPolicyType', choices=PolicyChoices, default=None,
                             help=clean_help('Purpose-grown biomass policy type. Default is None.'))

        parser.add_argument('-R', '--resultsDir', default='.',
                            help=clean_help('The parent directory holding the GCAM output workspaces'))

        parser.add_argument('-S', '--subdir', default='',
                             help=clean_help('Sub-directory for local-xml files, if any'))

        parser.add_argument('-t', '--fuelTag', default=DefaultTag,
                             help=clean_help("The fuel tag to generate constraints for. Default is %s" % DefaultTag))

        parser.add_argument('-T', '--policyType', choices=PolicyChoices, default=DefaultType,
                             help=clean_help('Type of policy to use for the fuel. Default is %s.' % DefaultType))

        parser.add_argument('-x', '--xmlOutputDir',
                             help=clean_help('''The directory into which to generate XML files.
                             Defaults to policy name in the current directory.'''))

        parser.add_argument('-y', '--years', default=DefaultYears,
                            help=clean_help('''Years to generate constraints for. Must be of the form
                            XXXX-YYYY. Default is "%s"''' % DefaultYears))
        return parser


    def run(self, args, tool):
        from ..constraints import genDeltaConstraints

        genDeltaConstraints(**vars(args))



'''
.. Copyright (c) 2019 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
'''
from .error import PygcamException, FileMissingError
from .log import getLogger
import os

_logger = getLogger(__name__)

_csvCache = {}

def readCachedCsv(filename, skiprows=1, cache=False):
    """
    Read a CSV file of the form generated by GCAM batch queries, i.e., skip one
    row and then read column headings and data. Optionally drop all years outside
    the `years` given. Optionally, linearly interpolate annual values between
    time-steps.

    :param filename: (str) the path to a CSV file
    :param skiprows: (int) the number of rows to skip before reading the data matrix
    :param cache: (bool) If True, file will be sought in, and saved to, a CSV cache.
       The "raw" file data is cached, so if called with different processing args,
       the same initial DataFrame is used, but it will be processed correctly.
    :return: (DataFrame) the data read in, processed as per arguments
    """
    import pandas as pd

    found = False

    if cache and filename in _csvCache:
        _logger.debug("Found %s in CSV cache", filename)
        df = _csvCache[filename]
        found = True

    else:
        try:
            _logger.debug("Reading %s", filename)
            df = pd.read_table(filename, sep=',', skiprows=skiprows, index_col=None)

        except IOError as e:
            raise FileMissingError(os.path.abspath(filename), e)

        except Exception as e:
            raise PygcamException('Error reading %s: %s' % (filename, e))

    # Cache a copy, unless it was already in the cache
    if cache and not found:
        _csvCache[filename] = df.copy()

    return df



from .building_plugin import BuildingCommand
from .buildingElec_plugin import BuildingElecCommand
from .transport_plugin import TransportCommand
from .chart_plugin import ChartCommand
from .config_plugin import ConfigCommand
from .diff_plugin import DiffCommand
from .gcam_plugin import GcamCommand
from .gui_plugin import GUICommand
from .init_plugin import InitCommand
from .mcs_plugin import MCSCommand
from .mi_plugin import ModelInterfaceCommand
from .new_plugin import NewProjectCommand
from .protect_plugin import ProtectLandCommand
from .query_plugin import QueryCommand
from .res_plugin import RESCommand
from .run_plugin import RunCommand
from .sandbox_plugin import SandboxCommand
from .setup_plugin import SetupCommand
from .compare_plugin import CompareCommand
from .zev_plugin import ZEVCommand
from.industry_plugin import IndustryCommand

BuiltinSubcommands = [BuildingCommand, BuildingElecCommand, ChartCommand, CompareCommand,
                      ConfigCommand, DiffCommand, GcamCommand, GUICommand, IndustryCommand,
                      InitCommand, MCSCommand, ModelInterfaceCommand, NewProjectCommand,
                      ProtectLandCommand, QueryCommand, RESCommand, RunCommand, SandboxCommand,
                      SetupCommand, TransportCommand, ZEVCommand]



'''
.. The "gt" (gcamtool) commandline program

.. Copyright (c) 2016-2017 Richard Plevin and UC Regents
   See the https://opensource.org/licenses/MIT for license details.
'''
from __future__ import print_function
import argparse
from glob import glob
import os
import pipes
import re
from semver import VersionInfo
import subprocess
import sys

from .config import (pathjoin, getParam, getConfig, getParamAsBoolean, getParamAsFloat,
                     setParam, getSection, setSection, getSections, DEFAULT_SECTION,
                     usingMCS, savePathMap, parse_version_info, setInputFilesByVersion)
from .error import PygcamException, ProgramExecutionError, ConfigFileError, CommandlineError
from .log import getLogger, setLogLevels, configureLogs
from .signals import SignalException, catchSignals
from .subcommand import clean_help
from .version import VERSION
from .windows import IsWindows

PROGRAM = 'gt'

# For now, these are not offered as command-line options. Needs more testing.
# BioConstraintsCommand, DeltaConstraintsCommand,

class GcamTool(object):

    # plugin instances by command name
    _plugins = {}

    # cached plugin paths by command name
    _pluginPaths = {}

    @classmethod
    def getPlugin(cls, name):
        if name not in cls._plugins:
            cls._loadCachedPlugin(name)

        return cls._plugins.get(name, None)

    @classmethod
    def _loadCachedPlugin(cls, name):
        # see if it's already loaded
        path = cls._pluginPaths[name]
        tool = cls.getInstance()
        tool.loadPlugin(path)

    @classmethod
    def _cachePlugins(cls):
        '''
        Find all plugins via GCAM.PluginPath and create a dict
        of plugin pathnames keyed by command name so the plugin
        can be loaded on-demand.
        :return: none
        '''
        pluginDirs = cls._getPluginDirs()

        suffix = '_plugin.py'
        suffixLen = len(suffix)

        for d in pluginDirs:
            pattern = pathjoin(d, '*' + suffix)
            for path in glob(pattern):
                basename = os.path.basename(path)
                command = basename[:-suffixLen]
                cls._pluginPaths[command] = path

    _instance = None

    @classmethod
    def getInstance(cls, loadPlugins=True, reload=False):
        """
        Get the singleton instance of the GcamTool class.

        :param loadPlugins: (bool) If true, plugins are loaded (only
           when first allocated).
        :param reload: (bool) If true, a new GcamTool instance is
           created.
        :return: (GcamTool instance) the new or cached instance.
        """
        if reload:
            GcamTool._instance = None
            GcamTool._plugins = {}
            GcamTool._pluginPaths = {}

        if not GcamTool._instance:
            GcamTool._instance = cls(loadPlugins=loadPlugins)

        return GcamTool._instance

    @classmethod
    def pluginGroup(cls, groupName, namesOnly=False):
        objs = filter(lambda obj: obj.getGroup() == groupName, cls._plugins.values())
        result = sorted(map(lambda obj: obj.name, objs)) if namesOnly else list(objs)
        return result

    def __init__(self, loadPlugins=True, loadBuiltins=True):
        from .project import decacheVariables

        # address re-entry issue
        decacheVariables()

        self.mcsMode = ''
        self.shellArgs = None

        self.parser = self.subparsers = None
        self.addParsers()

        # load all built-in sub-commands
        if loadBuiltins:
            from .built_ins import BuiltinSubcommands
            for item in BuiltinSubcommands:
                self.instantiatePlugin(item)

        # If using MCS, load that set of built-ins, too
        if usingMCS():
            from .mcs.built_ins import MCSBuiltins
            for item in MCSBuiltins:
                self.instantiatePlugin(item)

        # Load external plug-ins found in plug-in path
        if loadPlugins:
            self._cachePlugins()

    def addParsers(self):
        self.parser = parser = argparse.ArgumentParser(prog=PROGRAM, prefix_chars='-+')

        parser.add_argument('+b', '--batch', action='store_true',
                            help=clean_help('''Run the commands by submitting a batch job using the command
                            given by config variable GCAM.BatchCommand. (Linux only)'''))

        parser.add_argument('+B', '--showBatch', action="store_true",
                            help=clean_help("Show the batch command to be run, but don't run it. (Linux only)"))

        parser.add_argument('+D', '--dirmap',
                            help=clean_help("""A comma-delimited sequence of colon-delimited directory names 
                            of the form "/some/host/path:/a/container/path, /host:cont, ...", 
                            mapping host dirs to their mount point in a docker container."""))

        parser.add_argument('+e', '--enviroVars',
                            help=clean_help('''Comma-delimited list of environment variable assignments to pass
                            to queued batch job, e.g., -E "FOO=1,BAR=2". (Linux only)'''))

        parser.add_argument('+j', '--jobName', default='gt',
                            help=clean_help('''Specify a name for the queued batch job. Default is "gt".
                            (Linux only)'''))

        logLevel = str(getParam('GCAM.LogLevel'))   # so not unicode
        parser.add_argument('+l', '--logLevel',
                            default=logLevel or 'notset',
                            help=clean_help('''Sets the log level for modules of the program. A default
                                log level can be set for the entire program, or individual 
                                modules can have levels set using the syntax 
                                "module:level, module:level,...", where the level names must be
                                one of {debug,info,warning,error,fatal} (case insensitive).'''))

        parser.add_argument('+L', '--logFile',
                            help=clean_help('''Sets the name of a log file for batch runs. Default is "gt-%%j.out"
                            where "%%j" (in SLURM) is the jobid. If the argument is not an absolute
                            pathname, it is treated as relative to the value of GCAM.LogDir.'''))

        parser.add_argument('+m', '--minutes', type=float, default=getParamAsFloat('GCAM.Minutes'),
                            help=clean_help('''Set the number of minutes to allocate for the queued batch job.
                            Overrides config parameter GCAM.Minutes. (Linux only)'''))

        parser.add_argument('+M', '--mcs', dest='mcsMode', choices=['trial','gensim'],
                            help=clean_help('''Used only when running gcamtool from pygcam-mcs.'''))

        parser.add_argument('+P', '--projectName', metavar='name', default=getParam('GCAM.DefaultProject'),
                            choices=sorted(getSections()),
                            help=clean_help('''The project name (the config file section to read from),
                            which defaults to the value of config variable GCAM.DefaultProject'''))

        parser.add_argument('+q', '--queueName', default=getParam('GCAM.DefaultQueue'),
                            help=clean_help('''Specify the name of the queue to which to submit the batch job.
                            Default is given by config variable GCAM.DefaultQueue. (Linux only)'''))

        parser.add_argument('+r', '--resources', default='',
                            help=clean_help('''Specify resources for the queued batch command. Can be a comma-delimited
                            list of assignments of the form NAME=value, e.g., -r 'pvmem=6GB'. (Linux only)'''))

        parser.add_argument('+s', '--set', dest='configVars', metavar='name=value', action='append', default=[],
                            help=clean_help('''Assign a value to override a configuration file parameter. For example,
                            to set batch commands to start after a prior job of the same name completes,
                            use --set "GCAM.OtherBatchArgs=-d singleton". Enclose the argument in quotes if
                            it contains spaces or other characters that would confuse the shell.
                            Use multiple --set flags and arguments to set multiple variables.'''))

        parser.add_argument('+v', '--verbose', action='store_true',
                            help=clean_help('''Show diagnostic output'''))

        parser.add_argument('--version', action='version', version=VERSION)   # goes to stderr, handled by argparse

        parser.add_argument('--VERSION', action='store_true')   # goes to stdout, but handled by gt

        self.subparsers = self.parser.add_subparsers(dest='subcommand', title='Subcommands',
                               description='''For help on subcommands, use the "-h" flag after the subcommand name''')

    def setMcsMode(self, mode):
        self.mcsMode = mode

    def getMcsMode(self):
        return self.mcsMode

    def instantiatePlugin(self, pluginClass):
        plugin = pluginClass(self.subparsers)
        self._plugins[plugin.name] = plugin

    @staticmethod
    def _getPluginDirs():
        pluginPath = getParam('GCAM.PluginPath')
        if not pluginPath:
            return []

        sep = os.path.pathsep           # ';' on Windows, ':' on Unix
        items = pluginPath.split(sep)

        return items

    def loadPlugin(self, path):
        """
        Load the plugin at `path`.

        :param path: (str) the pathname of a plugin file.
        :return: an instance of the ``SubcommandABC`` subclass defined in `path`
        """
        from .utils import loadModuleFromPath

        def getModObj(mod, name):
            return getattr(mod, name) if name in mod.__dict__ else None

        mod = loadModuleFromPath(path)

        pluginClass = getModObj(mod, 'PluginClass') or getModObj(mod, 'Plugin')
        if not pluginClass:
            raise PygcamException('Neither PluginClass nor class Plugin are defined in %s' % path)

        self.instantiatePlugin(pluginClass)

    def _loadRequiredPlugins(self, argv):
        # Create a dummy subparser to allow us to identify the requested
        # sub-command so we can load the module if necessary.
        parser = argparse.ArgumentParser(prog=PROGRAM, add_help=False, prefix_chars='-+')
        parser.add_argument('-h', '--help', action='store_true')
        parser.add_argument('+P', '--projectName', metavar='name')

        ns, otherArgs = parser.parse_known_args(args=argv)

        # For top-level help, or if no args, load all plugins
        # so the generated help messages includes all subcommands
        if ns.help or not otherArgs:
            for item in self._pluginPaths.keys():
                self._loadCachedPlugin(item)
        else:
            # Otherwise, load any referenced sub-command
            for command in self._pluginPaths.keys():
                if command in otherArgs:
                    self.getPlugin(command)

    def validateGcamVersion(self):
        from .gcam import getGcamVersion
        from semver import VersionInfo

        exeDir  = pathjoin(getParam('GCAM.RefWorkspace'), 'exe')
        exeName = getParam('GCAM.Executable')
        exePath = pathjoin(exeDir, exeName)

        if not os.path.lexists(exePath):
            raise ConfigFileError('GCAM executable "%s" was not found.' % exePath)

        # Starting with v4.3, gcam reports its version number
        versionCfg = parse_version_info()
        if versionCfg >= VersionInfo(4, 3, 0):
            versionFile = pathjoin(exeDir, '.version')

            # Check for cached version info
            if os.path.lexists(versionFile):
                with open(versionFile, 'r') as f:
                    versionNum = f.readline().strip()
            else:
                versionNum = getGcamVersion(exeDir)

                # cache version number so we don't have to run 'gcam.exe --versionID' every time
                with open(versionFile, 'w') as f:
                    f.write(versionNum + '\n')

            versionExe = parse_version_info(versionNum)
            if (versionCfg.major, versionCfg.minor) != (versionExe.major, versionExe.minor):
                # use only major.minor to identify GCAM version
                versionNum = "{}.{}".format(versionExe.major, versionExe.minor)
                setParam('GCAM.VersionNumber', versionNum)

                log = getLogger(__name__)
                log.warning("Setting GCAM.VersionNumber = %s to match GCAM version. (Set it in the config file to suppress this message.)", versionNum)

        setInputFilesByVersion()

    def run(self, args=None, argList=None):
        """
        Parse the script's arguments and invoke the run() method of the
        designated sub-command.

        :param args: an argparse.Namespace of parsed arguments
        :param argList: (list of str) argument list to parse (when called recursively)
        :return: none
        """
        from .utils import getRegionList

        assert args or argList, "GcamTool.run requires either args or argList"

        checkWindowsSymlinks()

        if argList is not None:         # might be called with empty list of subcmd args
            # called recursively
            self._loadRequiredPlugins(argList)
            args = self.parser.parse_args(args=argList)

        else:  # top-level call
            if args.batch:
                args.batch = False

            args.projectName = section = args.projectName or getParam('GCAM.DefaultProject')
            if section:
                 setSection(section)

            if args.subcommand != 'init':
                # After GCAM.DefaultProject is set, we can check gcam executable
                try:
                    self.validateGcamVersion()
                except ConfigFileError as e:
                    _logger = getLogger(__name__)
                    _logger.warning("%s", e)

                getRegionList()

            logLevel = args.logLevel or getParam('GCAM.LogLevel')
            if logLevel:
                setLogLevels(logLevel)

            configureLogs(force=True)

        # Get the sub-command and run it with the given args
        obj = self.getPlugin(args.subcommand)

        obj.run(args, self)

    # Extracted from runBatch() to be callable from the run sub-command
    # to distribute scenario runs.
    @staticmethod
    def runBatch2(shellArgs, jobName='gt', queueName=None, logFile=None, minutes=None,
                  dependsOn=None, run=True):
        from .utils import mkdirs

        _logger = getLogger(__name__)

        queueName = queueName or getParam('GCAM.DefaultQueue')
        logFile   = logFile   or getParam('GCAM.BatchLogFile', raw=False)
        minutes   = minutes   or getParamAsFloat('GCAM.Minutes')

        batchSystem = getParam('GCAM.BatchSystem')
        known = ('SLURM', 'LSF', 'PBS')
        if batchSystem not in known:
            raise ConfigFileError('GCAM.Scheduler value (%s) is not recognized. Must be one of %s.' % (batchSystem, known))

        # The LSF scheduler needs HH:MM; SLURM needs HH:MM:SS
        format = "%02d:%02d" if batchSystem == 'LSF' else "%02d:%02d:00"
        walltime  = format % (minutes // 60, minutes % 60)

        if logFile:
            logDir = getParam('GCAM.BatchLogDir')
            logFile = os.path.normpath(pathjoin(logDir, logFile))
            mkdirs(os.path.dirname(logFile))

        # TBD: make this an expression eval'd with s.format(jobID=dependsOn)
        # TBD: to support other syntaxes
        format = "-w 'done(%s)'" if batchSystem == 'LSF' else "-d afterok:%s"
        dependencies = format % dependsOn if dependsOn else ''

        scriptCommand = "gt " + ' '.join(shellArgs)

        # This dictionary is applied to the string value of GCAM.BatchCommand, via
        # the str.format method, which must specify options using any of the keys.
        batchArgs = {'scriptFile': scriptCommand,
                     'dependencies' : dependencies,
                     'logFile'   : logFile,
                     'minutes'   : minutes,
                     'walltime'  : walltime,
                     'queueName' : queueName,
                     'partition' : queueName,   # synonym for queue name
                     'jobName'   : jobName}

        batchCmd = getParam('GCAM.BatchCommand')

        try:
            command = batchCmd.format(**batchArgs)

        except KeyError as e:
            raise ConfigFileError('Badly formatted batch command (%s) in config file: %s', batchCmd, e)

        if not run:
            print(command)
            return

        _logger.info('Running: %s', command)
        try:
            jobStr = subprocess.check_output(command, shell=True).decode('utf-8')
            result = re.search('\d+', jobStr)
            jobId = int(result.group(0)) if result else -1
            return jobId

        except subprocess.CalledProcessError as e:
            raise ProgramExecutionError(command, e.returncode)

        except Exception as e:
            raise PygcamException("Error running command '%s': %s" % (command, e))

    def runBatch(self, shellArgs, run=True):
        import platform

        system = platform.system()
        if system in ['Windows']: # , 'Darwin']:
            system = 'Mac OS X' if system == 'Darwin' else system
            raise CommandlineError('Batch commands are not supported on %s' % system)

        shellArgs = [pipes.quote(arg) for arg in shellArgs]
        args = self.parser.parse_args(args=shellArgs)

        return self.runBatch2(shellArgs, jobName=args.jobName, queueName=args.queueName,
                              logFile=args.logFile, minutes=args.minutes, run=run)

def _getMainParser():
    '''
    Used only to generate documentation by sphinx' argparse, in which case
    we don't generate documentation for project-specific plugins.
    '''
    getConfig(allowMissing=True)
    tool = GcamTool.getInstance(loadPlugins=False)
    return tool.parser


def checkWindowsSymlinks():
    '''
    If running on Windows and GCAM.CopyAllFiles is not set, and
    we fail to create a test symlink, set GCAM.CopyAllFiles to True.
    '''
    from .temp_file import getTempFile

    if IsWindows and not getParamAsBoolean('GCAM.CopyAllFiles'):
        src = getTempFile()
        dst = getTempFile()

        try:
            os.symlink(src, dst)
        except:
            _logger = getLogger(__name__)
            _logger.info('No symlink permission; setting GCAM.CopyAllFiles = True')
            setParam('GCAM.CopyAllFiles', 'True')

def _setDefaultProject(argv):
    parser = argparse.ArgumentParser(prog=PROGRAM, add_help=False, prefix_chars='-+')
    parser.add_argument('+P', '--projectName', metavar='name')

    ns, _otherArgs = parser.parse_known_args(args=argv)

    section = ns.projectName
    if section:
        setParam('GCAM.DefaultProject', section, section=DEFAULT_SECTION)
        setSection(section)

    # Set the data dir based on the version of the model used in this project
    version = parse_version_info()

    v_5_1_0 = VersionInfo(5, 1, 0)

    dataDir = "gcamdata" if version >= v_5_1_0 else "gcam-data-system"
    setParam('GCAM.DataDir', dataDir, section=section)

    # ModelInterface was also relocated in v5.1, so we compute the path when setting the project
    subdir = 'output/modelinterface' if version >= v_5_1_0 else 'input/gcam-data-system/_common/ModelInterface/src'
    setParam('GCAM.MI.Subdir', subdir )


def _saveDirMap():
    dirMapFile = os.getenv('DIRMAP_PATH')
    if not dirMapFile:
        return

    with open(dirMapFile) as f:
        dirMap = f.read()

    savePathMap(dirMap)

# This parser handles only --VERSION flag.
def _showVersion(argv):
    parser = argparse.ArgumentParser(prog=PROGRAM, add_help=False, prefix_chars='-+')
    parser.add_argument('--VERSION', action='store_true')

    ns, otherArgs = parser.parse_known_args(args=argv)

    if ns.VERSION:
        print(VERSION)
        sys.exit(0)

def _main(argv=None):
    from .config import userConfigPath

    configPath = userConfigPath()
    if not os.path.lexists(configPath) or os.stat(configPath).st_size == 0:
        argSet = set(argv or sys.argv)
        options = {'init', '-h', '--help', '--version'}
        if argSet.intersection(options):
            # create empty config file just so we can run the "init" sub-command, or help/version options
            open(configPath, 'w').close()
        else:
            raise CommandlineError('\n***\n*** Missing or empty pygcam configuration file %s. Run "gt init" to create it.\n***\n' % configPath)

    getConfig()

    _showVersion(argv)
    _saveDirMap()

    configureLogs()

    _setDefaultProject(argv)

    tool = GcamTool.getInstance()
    tool._loadRequiredPlugins(argv)

    # This parser handles only --batch, --showBatch, --projectName, --set, and --mcs
    # args. If --batch is given, we need to create a script and call the
    # GCAM.BatchCommand on it. We grab --projectName so we can set PluginPath by project.
    parser = argparse.ArgumentParser(prog=PROGRAM, add_help=False, prefix_chars='-+')

    parser.add_argument('+b', '--batch', action='store_true')
    parser.add_argument('+B', '--showBatch', action="store_true")
    parser.add_argument('+P', '--projectName', dest='projectName', metavar='name')
    parser.add_argument('+s', '--set', dest='configVars', action='append', default=[])
    parser.add_argument('+M', '--mcs', dest='mcsMode', choices=['trial','gensim'])

    ns, otherArgs = parser.parse_known_args(args=argv)

    tool.setMcsMode(ns.mcsMode)

    # Set specified config vars
    for arg in ns.configVars:
        if not '=' in arg:
            raise CommandlineError('+s requires an argument of the form variable=value, got "%s"' % arg)

        name, value = arg.split('=')
        setParam(name, value)

    # showBatch => don't run batch command, but implies --batch
    if ns.showBatch:
        ns.batch = True

    # Catch signals to allow cleanup of TempFile instances, e.g., on ^C
    catchSignals()

    if ns.batch:
        run = not ns.showBatch
        if ns.projectName:        # add these back in for the batch script
            otherArgs = ['+P', ns.projectName] + otherArgs

        tool.runBatch(otherArgs, run=run)
    else:
        tool.shellArgs = otherArgs  # save for project run method to use in "distribute" mode
        args = tool.parser.parse_args(args=otherArgs)
        tool.run(args=args)


def main(argv=None, raiseError=False):
    try:
        _main(argv)
        return 0

    except CommandlineError as e:
        print(e)

    except SignalException as e:
        if raiseError:
            raise

        _logger = getLogger(__name__)
        _logger.error("%s: %s" % (PROGRAM, e))
        return e.signum

    except Exception as e:
        if raiseError:
            raise

        print("%s failed: %s" % (PROGRAM, e))

        if not getSection() or getParamAsBoolean('GCAM.ShowStackTrace'):
            import traceback
            traceback.print_exc()

    finally:
        from .temp_file import TempFile

        # Delete any temporary files that were created
        TempFile.deleteAll()

    return 1



"""
.. Support for generating renewable energy standards described in an XML file.

.. codeauthor:: Rich Plevin <rich@plevin.com>

.. Copyright (c) 2020 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
"""
from ..subcommand import SubcommandABC, clean_help

class ZEVCommand(SubcommandABC):
    def __init__(self, subparsers):
        kwargs = {'help': clean_help('''Generate an CSV template file that can be used to implement a
            ZEV policy on the sectors, transSubsectors and technologies as described on the command-line.''')}

        super(ZEVCommand, self).__init__('zev', subparsers, kwargs, group='project')

    def addArgs(self, parser):
        parser.add_argument('-i', '--include', action='append', default=None,
                            help=clean_help('''A colon (":") delimited list of comma-delimited sectors, 
                                tranSubsectors, and technologies to include in the CSV template file.
                                Example: "--include trn_pass_road_LDV_4W::BEV,FCEV" means include only two 
                                technologies (BEV,FCEV), but for any tranSubsector under the specified sector. 
                                Multiple -I arguments are allowed.'''))

        defaultCSV = 'zev_policy.csv'
        parser.add_argument('-o', '--outputCSV', default=defaultCSV,
                            help=clean_help('''The directory into which to write the generated CSV template.
                                    Default is "{}". If set to a relative pathname, it is assumed to 
                                    be relative to %%(GCAM.ProjectDir)s/etc.'''.format(defaultCSV)))

        parser.add_argument('-r', '--regions', default=None,
                            help=clean_help('''A comma-delimited list of regions to include in the generated template. 
                             By default all regions are included. '''))

        parser.add_argument('-S', '--scenario', default=None,
                            help=clean_help('''The name of the scenario for which to generate the policy 
                                    implementation XML file. Required if no argument is given
                                    to the "-o/--outputXML" flag, or if the argument is a
                                    relative pathname.'''))

        parser.add_argument('-t', '--tag', default='transportation',
                            help=clean_help('''The config file tag identifying the transportation file to operate on.'''))

        parser.add_argument('-u', '--GCAM-USA', action="store_true",
                            help=clean_help('''If set, produce output compatible with GCAM-USA regions.'''))

        parser.add_argument('-y', '--years', default='2015-2100',
                            help=clean_help('''A hyphen-separated range of timestep years to include in the generated template.
                            Default is "2015-2100"'''))

        return parser

    def run(self, args, tool):
        from ..ZEVPolicy import zevPolicyMain
        zevPolicyMain(args)



from pygcam.log import getLogger

_logger = getLogger(__name__)

# Templates
Policy_year_template = """
      <policy-portfolio-standard name="{policy}">
        <market>{market}</market>
        <policyType>RES</policyType>
        <constraint year="{year}">1</constraint>
      </policy-portfolio-standard>"""

# stub-tech wrapped in <supplysector name="x"><subsector name="y"><stub-technology...
Supply_year_template = """
            <period year="{year}">
              <minicam-energy-input name="{policy}">
                <coefficient>{coefficient}</coefficient>
                <market-name>{market}</market-name>
              </minicam-energy-input>{secondaryInput}
            </period>"""

SecondaryInput_template = """
              <res-secondary-output name="{policy}">
                <output-ratio>1</output-ratio>
                <pMultiplier>1</pMultiplier>
              </res-secondary-output>"""

Subsector_template = """
        <subsector name="{subsector}">{technologyElts}
        </subsector>"""
        
Technology_template = """
          <stub-technology name="{technology}">{periodElts}
          </stub-technology>"""

Supply_template = """
      <supplysector name="{sector}">{subsectorElts}
      </supplysector>"""

Region_template = """
    <region name="{region}">{policyElts}
      {supplysectorElts}
    </region>"""

Policy_file_template = """<?xml version="1.0" encoding="UTF-8"?>
<scenario>
  <world>{regionElts}
  </world>
</scenario>
"""

def emit_policy_year(market, year, policy):
    return Policy_year_template.format(market=market, year=year, policy=policy)

def emit_supply_year(market, year, policy, coefficient, secondaryInput):
    return Supply_year_template.format(market=market, year=year, policy=policy,
                                       coefficient=coefficient,
                                       secondaryInput=secondaryInput)

def emit_subsector(subsector, technologyEltList):
    technologyElts = ''.join(technologyEltList)
    return Subsector_template.format(subsector=subsector, technologyElts=technologyElts)

def emit_technology(technology,periodEltList):
    periodElts = ''.join(periodEltList)
    return Technology_template.format(technology=technology,
                                     periodElts=periodElts)

def emit_supply(sector, subsectorEltList):
    subsectorElts = ''.join(subsectorEltList)
    return Supply_template.format(sector=sector, subsectorElts=subsectorElts)

def emit_region(region, policyEltList, supplysectorEltList):
    policyElts = ''.join(policyEltList)
    supplysectorElts = ''.join(supplysectorEltList)
    return Region_template.format(region=region, policyElts=policyElts,
                                  supplysectorElts=supplysectorElts)

def emit_policy_file(regionEltList):
    regionElts = ''.join(regionEltList)
    return Policy_file_template.format(regionElts=regionElts)

def generate_building_elec_xml(csv_path, xml_path):
    import pandas as pd
    from collections import OrderedDict

    _logger.debug("Reading '%s'", csv_path)
    df = pd.read_csv(csv_path, index_col=None, skiprows=0)

    year_cols = [col for col in df.columns if col.isdigit()]
    sort_cols = ['region', 'sector', 'subsector','technology']
    df.sort_values(by=sort_cols, inplace=True)

    region_dict = {}
    market_dict = {}    # market name by region

    def subdict(d, key):
        if key not in d:
            d[key] = {}
        return d[key]

    # create hierarchical dict struct from dataframe
    for (idx, row) in df.iterrows():
        market_dict[row.region] = row.region
        sector_dict    = subdict(region_dict, row.region)
        subsector_dict = subdict(sector_dict, row.sector)
        technology_dict = subdict(subsector_dict, row.subsector)
        period_dict    = subdict(technology_dict, row.technology)
        for col in year_cols:
            period_dict[col] = row[col]

    _logger.info("Writing '%s'", xml_path)
    with open(xml_path, 'w') as f:

        policy_elts = OrderedDict()
        region_elts = []
        for (region, sector_dict) in region_dict.items():
            market = market_dict[region]

            supplysector_elts = []
            for (sector, subsector_dict) in sector_dict.items():

                subsector_elts = []
                for (subsector, technology_dict) in subsector_dict.items():
                    
                    technology_elts=[]
                    for (technology, period_dict) in technology_dict.items():
                        
                        period_elts = []
                        for (year, coefficient) in period_dict.items():
                            policy = "BuildingElec-{}-{}-{}".format(market, technology, year)
                            secondaryInput = SecondaryInput_template.format(policy=policy) if subsector == 'electricity' else ''
                            period_elts.append(emit_supply_year(market, year, policy, coefficient, secondaryInput))    

                            policy_elts[policy] = emit_policy_year(market, year, policy)   # avoids duplicates since policy name is reused
                            
                        technology_elts.append(emit_technology(technology,period_elts))

                    subsector_elts.append(emit_subsector(subsector, technology_elts))

                supplysector_elts.append(emit_supply(sector, subsector_elts))

            region_elts.append(emit_region(region, policy_elts.values(), supplysector_elts))

        f.write(emit_policy_file(region_elts))

if __name__ == '__main__':
    dir_path = '/tmp/'
    csv_path = dir_path + 'building-elec-template.csv'
    xml_path = dir_path + 'building-elec-policy.xml'

    generate_building_elec_xml(csv_path, xml_path)
